#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>

#include "argument.h"
#include "text.h"
#include "lib.h"

static char *ifname = NULL;
static int bufsize  = PKT_BUFFER_SIZE_DEFAULT;
static int waitusec = 0;
static int complete = ARGUMENT_FLAG_OFF;
static int interval = ARGUMENT_FLAG_ON;

static Argument args[] = {
  { "-b", ARGUMENT_TYPE_INTEGER , &bufsize  },
  { "-i", ARGUMENT_TYPE_STRING  , &ifname   },
  { "-w", ARGUMENT_TYPE_INTEGER , &waitusec },
  { "-c", ARGUMENT_TYPE_FLAG_ON , &complete },
  { "-f", ARGUMENT_TYPE_FLAG_OFF, &interval },
  { NULL, ARGUMENT_TYPE_NONE    , NULL      },
};

static int timecmp(struct timeval *t0, struct timeval *t1)
{
  if (t0->tv_sec  > t1->tv_sec ) return  1;
  if (t0->tv_sec  < t1->tv_sec ) return -1;
  if (t0->tv_usec > t1->tv_usec) return  1;
  if (t0->tv_usec < t1->tv_usec) return -1;
  return 0;
}

static int timesub(struct timeval *t0, struct timeval *t1)
{
  if (timecmp(t0, t1) < 0)
    return -1;
  while (t0->tv_usec < t1->tv_usec) {
    t0->tv_sec--;
    t0->tv_usec += 1000000;
  }
  t0->tv_sec  -= t1->tv_sec;
  t0->tv_usec -= t1->tv_usec;
  return 0;
}

static int terminated = 0;
static void sigint_handler(int value)
{
  terminated = 1;
}

int main(int argc, char *argv[])
{
  unsigned long flags = 0;
  int fd, size;
  char *buffer;
  struct timeval starttime, firsttime, nowtime, tm;
  int first = 1;

  argument_read(&argc, argv, args);
  if (ifname == NULL)
    error_exit("Unknown interface.\n");
  if (complete) flags |= PKT_SEND_FLAG_COMPLETE;
  if (interval) flags |= PKT_SEND_FLAG_INTERVAL;

  fd = pkthandler.open_send(ifname, flags);

  buffer = malloc(bufsize);
  if (buffer == NULL)
    error_exit("Out of memory.\n");

  while (!terminated) {
    size = pkt_text_read(stdin, buffer, bufsize, NULL, NULL, &tm);
    if (size < 0)
      break;
    if (size == bufsize)
      error_exit("Out of buffer.\n");

    if (flags & PKT_SEND_FLAG_INTERVAL) {
      if (first) {
	firsttime.tv_sec  = tm.tv_sec;
	firsttime.tv_usec = tm.tv_usec;
	gettimeofday(&starttime, NULL);
	first = 0;
      } else {
	gettimeofday(&nowtime, NULL);
	if (timesub(&tm, &firsttime) == 0)
	  if (timesub(&nowtime, &starttime) == 0)
	    if (timesub(&tm, &nowtime) == 0)
	      select(0, NULL, NULL, NULL, &tm);
      }
    }

    signal(SIGINT , sigint_handler);
    signal(SIGTERM, sigint_handler);
    pkthandler.send(fd, buffer, size);
    signal(SIGINT , SIG_DFL);
    signal(SIGTERM, SIG_DFL);

    if (waitusec > 0)
      usleep(waitusec);
  }

  free(buffer);

  close(fd);

  return 0;
}
